from collections import Counter


class MagicDict(dict):

    def __getitem__(self, item):
        try:
            return dict.__getitem__(self, item)
        except KeyError:
            # create a self instance
            value = self[item] = type(self)()
            return value

    @staticmethod
    def reverse(dic):
        """
        Preserve all repeated value-keys when a Dict is reversed.

        Parameters
        ----------
        dic: dict
            A dict where several keys have the same values.

        Returns
        -------
        Reversed dict of the given dict, but preserve all key-values.

        Example
        -------
        dx = {  1: 'a',
                2: 'a',
                3: 'a',
                4: 'b' }
        reversedx = {   'a': [1, 2, 3],
                        'b': 4 }
        """
        d1 = dict(zip(dic.values(), [[] for i in range(len(dic))]))
        d2 = dict([
            (y, d1[y].append(x))
            if y in
            [w for (w, f) in Counter(dic.values()).items() if f > 1]
            else (y, x)
            for (x, y) in dic.items()])
        reversdict = dict([(x, d1[x]) if len(d1[x]) != 0
                           else (x, d2[x])
                           for x in d1.keys()])
        return reversdict



from .piop import read_file, read_lines, read_json, read_yaml, read_csv
from .piop import write_file, write_json, check_dir
from .pcut import cut_sentence, cut_zhchar, cut_part, combine_bucket
from .pcut import psent, psubsent

from .piop import Reader, Dict
from .ptxt import Regex, Text, Length
from .pnorm import NumNorm
from .penh import TokenLevelSampler, SentenceLevelSampler
from .pmag import MagicDict
from .stopwords import StopWords

from .utils import pstr
from .stopwords import chinese_stopwords, english_stopwords


num_norm = NumNorm()
reg = Regex()
reader = Reader()
tlsampler = TokenLevelSampler()
slsampler = SentenceLevelSampler()


__title__ = 'pnlp'
__version__ = '0.3.5'
__author__ = 'Yam'
__license__ = 'Apache-2.0'
__copyright__ = 'Copyright 2019, 2020 Yam'
__all__ = ['Reader',
           'Text', 'Regex', 'Length',
           'MagicDict',
           'NumNorm',
           'StopWords',
           'TokenLevelSampler', 'SentenceLevelSampler'
           ]



# from __future__ import annotations

from addict import Dict
import json
import os
import csv
import pathlib
import yaml


class Reader:

    def __init__(self, pattern: str = "*.*"):
        self.pattern = pattern

    def __repr__(self) -> str:
        return "Reader(pattern=%r)" % (self.pattern)

    @staticmethod
    def gen_files(dirname: str, pattern: str = '*.*'):
        """
        Find all filenames in a directory tree that match the filepattern.
        If filepath is a file, yield the filepath directly.
        """
        if os.path.isfile(dirname):
            fpath = pathlib.Path(dirname)
            yield fpath
        for fpath in pathlib.Path(dirname).rglob(pattern):
            yield fpath

    @staticmethod
    def gen_articles(fpaths: list):
        for fpath in fpaths:
            with open(fpath, encoding='utf8') as f:
                article = Dict()
                article.fname = fpath.name
                article.f = f
                yield article

    @staticmethod
    def gen_flines(articles: list):
        """
        Process each file to lines when io.TextIOWrapper is given.
        """
        for article in articles:
            lid = 0
            for line_content in article.f:
                line_content = line_content.strip()
                if len(line_content) == 0:
                    continue
                line = Dict()
                line.lid = lid
                line.fname = article.fname
                line.text = line_content
                lid += 1
                yield line

    @staticmethod
    def gen_plines(fpath: str):
        """
        Process each file to lines when fpath is given.
        """
        with open(fpath, encoding='utf8') as f:
            for line in f:
                line = line.strip()
                if len(line) == 0:
                    continue
                yield line

    def __call__(self, dirname: str):
        fpaths = self.gen_files(dirname, self.pattern)
        articles = self.gen_articles(fpaths)
        flines = self.gen_flines(articles)
        for line in flines:
            yield line


def read_file(fpath: str, **kwargs) -> str:
    """
    Read file from file path.

    Parameters
    -----------
    fpath: str
        File path.
    kwargs: optional
        Other `open` support params. 

    Returns
    --------
        data string of the file.
    """
    with open(fpath, **kwargs) as f:
        data = f.read()
    return data


def read_lines(fpath: str, **kwargs) -> list:
    """
    Read file with `open` from file path.

    Parameters
    ----------
    fpath: str
        File path.
    kwargs: optional
        Other `open` support params. 

    Returns
    -------
    Lines of the file.

    Notes
    -----
    Blank line is ignored.
    """
    res = []
    with open(fpath, **kwargs) as f:
        for line in f:
            line = line.strip()
            if len(line) == 0:
                continue
            res.append(line)
    return res


def read_csv(fpath: str, delimiter: str = ","):
    data = []
    with open(fpath, 'r') as f:
        fcsv = csv.reader(f, delimiter=delimiter)
        for row in fcsv:
            data.append(row)
    return data


def read_json(fpath: str, **kwargs):
    with open(fpath, 'r') as fin:
        data = json.load(fin, **kwargs)
    return data


def read_yaml(fpath: str):
    with open(fpath, 'r') as fin:
        data = yaml.load(fin, Loader=yaml.SafeLoader)
    return data


def write_json(fpath: str, data, **kwargs):
    fout = open(fpath, 'w')
    json.dump(data, fout, **kwargs)
    fout.close()


def write_file(fpath: str, data, **kwargs):
    with open(fpath, 'w', **kwargs) as fout:
        for line in data:
            fout.write(line + "\n")


def check_dir(dirname: str):
    if os.path.exists(dirname):
        pass
    else:
        os.makedirs(dirname)



from dataclasses import dataclass
from typing import TypeVar


T = TypeVar('T', str, float, int)

ZH_NUM = {
    '〇': 0, '一': 1, '二': 2, '三': 3, '四': 4, '五': 5,
    '六': 6, '七': 7, '八': 8, '九': 9, '零': 0,
    '壹': 1, '贰': 2, '叁': 3, '肆': 4, '伍': 5,
    '陆': 6, '柒': 7, '捌': 8, '玖': 9, '貮': 2, '两': 2,
}

ZH_UNIT = {
    '十': 10,
    '拾': 10,
    '百': 100,
    '佰': 100,
    '千': 1000,
    '仟': 1000,
    '万': 10000,
    '萬': 10000,
    '亿': 100000000,
    '億': 100000000,
    '兆': 10000000000000,
}


ARB_NUM = {
    0: "零",
    1: "一",
    2: "二",
    3: "三",
    4: "四",
    5: "五",
    6: "六",
    7: "七",
    8: "八",
    9: "九",
    10: "十",
    100: "百",
    1000: "千",
    10000: "万",
    100000000: "亿",
    10000000000000: "兆"
}

ZH2MONEY = {
    "一": "壹",
    "二": "贰",
    "三": "叁",
    "四": "肆",
    "五": "伍",
    "六": "陆",
    "七": "柒",
    "八": "捌",
    "九": "玖",
    "十": "拾",
    "百": "佰",
    "千": "仟",
    "万": "萬",
    "亿": "億"
}


class pnumstr(str):

    def to_money(self):
        for c in self:
            mc = ZH2MONEY.get(c)
            if mc:
                self = self.replace(c, mc)
        return self


@dataclass
class NumNorm:
    """
    Chinese_to_Arabic
    modifed from https://github.com/bamtercelboo/corpus_process_script/blob/master/cn_to_arabic/cn_to_arabic.py
    """
    @staticmethod
    def num_len(num: int) -> int:
        if num == 0:
            return 1
        if num < 0:
            num = -num
        i = 0
        while num != 0:
            num //= 10
            i += 1
        return i

    def num2zh(self, num: int) -> str:
        def get_base(num):
            zh = ARB_NUM.get(num)
            if num < 10:
                return zh
            else:
                return "一" + zh

        def get_less_than_10w(num):
            res = ""
            while num != 0:
                if num < 10:
                    res += ARB_NUM.get(num)
                    break
                length = NumNorm.num_len(num)
                divider = 10 ** (length - 1)
                high = num // divider
                res += ARB_NUM.get(high)
                res += ARB_NUM.get(divider)
                num = num % divider
                new_len = NumNorm.num_len(num)
                if length - new_len > 1 and num != 0:
                    res += "零"
            return res

        def get_interval(num: int, lower: int, unit: str):
            res = ""
            length = NumNorm.num_len(num)
            divider = lower / 10
            high = num // divider
            res = get_less_than_10w(high)
            high_len = NumNorm.num_len(high)
            res += unit
            num -= high * divider
            new_len = NumNorm.num_len(num)
            if length - high_len - new_len > 0 and num != 0:
                res += "零"
            return res, num

        def get_10w_to_1y(num):
            res, num = get_interval(num, 10**5, "万")
            if 0 < num < 100000:
                res += get_less_than_10w(num)
            return res

        def get_1y_to_1z(num):
            res, num = get_interval(num, 10**9, "亿")
            if 0 < num < 100000000:
                res += get_10w_to_1y(num)
            return res

        if num in ARB_NUM:
            result = get_base(num)
            return pnumstr(result)
        # 十万
        if num < 10**5:
            result = get_less_than_10w(num)
        # 一亿
        elif num < 10**8:
            result = get_10w_to_1y(num)
        # 一兆
        elif num < 10**13:
            result = get_1y_to_1z(num)
        else:
            result = "超大"
        return pnumstr(result)

    def zh2num(self, zh: str) -> T:
        unit = 0
        digit_list = []
        for zhdigit in reversed(zh):
            if zhdigit in ZH_UNIT:
                unit = ZH_UNIT.get(zhdigit)
                if unit == 10000 or unit == 100000000:
                    digit_list.append(unit)
                    unit = 1
            else:
                digit = ZH_NUM.get(zhdigit)
                if unit:
                    digit *= unit
                    unit = 0
                digit_list.append(digit)
        if unit == 10:
            digit_list.append(10)
        val, tmp = 0, 0
        for x in reversed(digit_list):
            if x == 10000 or x == 100000000:
                val += tmp * x
                tmp = 0
            else:
                tmp += x
        val += tmp
        if val == 0 and zh != "零":
            return zh
        else:
            return val



from addict import Dict
import re


class Regex:

    """
    All kinds of Regular patterns.
    """

    patnames = ["chi", "pun",
                "whi", "nwh",
                "wnb", "nwn",
                "eng", "num",
                "pic", "lnk", "emj"]

    pun_zh = r"，。；、？！：“”‘’（）「」『』〔〕【】《》〈〉…——\-—～~·"
    pun_en = r",.;?!\(\)\[\]\{\}<>_"

    @property
    def pchi(self):
        """
        Chinese char pattern.
        """
        _pchi = re.compile(r'[\u4E00-\u9FD5]+')  # from jieba
        return _pchi

    @property
    def ppun(self):
        """
        Punctuation pattern.
        """
        _ppun = re.compile(rf'[{self.pun_en + self.pun_zh}]+')
        return _ppun

    @property
    def pwhi(self):
        """
        White space pattern.
        """
        _pwhi = re.compile(r'\s+')
        return _pwhi

    @property
    def pnwh(self):
        """
        Non-white space pattern.
        """
        _pnwh = re.compile(r'\S+')
        return _pnwh

    @property
    def pwnb(self):
        """
        Word and num pattern.
        """
        _pwnb = re.compile(r'\w+')
        return _pwnb

    @property
    def pnwn(self):
        """
        Non-alphanumeric char pattern.
        """
        _pnwn = re.compile(r'\W+')
        return _pnwn

    @property
    def peng(self):
        """
        English char pattern.
        """
        _peng = re.compile(r'[a-zA-Z]+')
        return _peng

    @property
    def pnum(self):
        """
        Number pattern.

        Example
        -------
        2, +2, -2, 2.1, -2.2, 1/5, 2:3, -2/5, 2%, 2.5%
        """
        _pnum = re.compile(r'''
                    [+-.]?\d+[.:/]?[\d%]+
                    |
                    [+-.]?\d+(?!\.\w+)
                    ''', re.UNICODE | re.VERBOSE)
        return _pnum

    @property
    def ppic(self):
        """
        Picture pattern.
        """
        _ppic = re.compile(r'''
                    !\[.*?\]\(.*?\.?(jpeg|png|jpg|gif)?\)
                    |
                    https?:\/\/(www\.)?[-a-zA-Z0-9@:%._\+~#=]{0,256}\.(jpeg|png|jpg|gif)
                    ''', re.UNICODE | re.VERBOSE)
        return _ppic

    @property
    def plnk(self):
        """
        Link pattern.
        """
        _plink = re.compile(r'''
                    \[.+?\]\(https?:\/\/(www\.)?[-a-zA-Z0-9@:%._\+~#=]{0,256}\.[a-z]{2,6}\b([-a-zA-Z0-9@:%_\+.~#?&//=]*)\)
                    |
                    https?:\/\/(www\.)?[-a-zA-Z0-9@:%._\+~#=]{0,256}\.[a-z]{2,6}\b([-a-zA-Z0-9@:%_\+.~#?&//=]*)
                    |
                    (https?:\/\/)?www\.[-a-zA-Z0-9@:%._\+~#=]{0,256}\.[a-z]{2,6}\b([-a-zA-Z0-9@:%_\+.~#?&//=]*)
                    ''', re.UNICODE | re.VERBOSE)
        return _plink

    @property
    def pemj(self):
        """
        Emoj pattern.
        """
        _pemj = re.compile(
            r'['u'\U0001F300-\U0001F64F'
                u'\U0001F680-\U0001F6FF'
                u'\u2600-\u2B55]+')
        return _pemj

    @property
    def patdict(self):
        """
        All patterns.
        """
        patterns = [self.pchi, self.ppun,
                    self.pwhi, self.pnwh,
                    self.pwnb, self.pnwn,
                    self.peng, self.pnum,
                    self.ppic, self.plnk, self.pemj]
        _patdict = dict(zip(self.patnames, patterns))
        return _patdict


class Text(Regex):

    """
    Text clean, extract and length.

    Parameters
    ----------
    pat: list of patterns
        Support custom re.Pattern. 
        Default is re.compile(r'.+').
        Other str pattern is built-in, including:
        'chi': Chinese character
        'pun': Punctuations
        'whi': White space
        'nwh': Non White space
        'wnb': Word and number
        'nwn': Non word and number
        'eng': English character
        'num': Number
        'pic': Pictures
        'lnk': Links
        'emj': Emojis

    Returns
    -------
    A text object.

    Notes
    ------
    The pattern order is important. The front pattern will be excute earlier.
    """

    def __init__(self, pattern_list: list = []):
        self.pats = []
        for pat in pattern_list:
            if isinstance(pat, str):
                built_in_pat = self.patdict.get(pat)
                if built_in_pat:
                    self.pats.append(built_in_pat)
                else:
                    raise ValueError(
                        "pnlp: {} \
                        is not a valid built-in pattern.".format(pat))
            elif isinstance(pat, re.Pattern):
                self.pats.append(pat)
            else:
                raise ValueError(
                    "pnlp: {} is not a valid RE pattern.".format(pat))

    def __repr__(self) -> str:
        return "Text(pattern=%r)" % str(self.pat)

    def clean(self, text: str):
        """
        Clean text with givening pattern.

        Returns
        -------
        Cleaned text.
        """
        for pat in self.pats:
            text = pat.sub("", text)
        return text

    def extract(self, text: str):
        """
        Extract pattern-matching items.

        Returns
        -------
        Extracted items and their locations.
        """
        mats, locs = [], []
        for pat in self.pats:
            for mat in pat.finditer(text):
                mats.append(mat.group())
                locs.append(mat.span())
        ext = Dict()
        ext.text = "".join(mats)
        ext.mats = mats
        ext.locs = locs
        return ext


class Length(Regex):

    def __init__(self, text: str):
        self.text = text

    def _len(self, pat):
        lst = pat.findall(self.text)
        return len("".join(lst))

    @property
    def len_all(self):
        """
        Length of all characters.
        """
        return len(self.text)

    @property
    def len_nwh(self):
        """
        Length of non-white characters.
        """
        return self._len(self.pnwh)

    @property
    def len_chi(self):
        """
        Length of pure Chinese characters.
        """
        return self._len(self.pchi)

    @property
    def len_wnb(self):
        """
        Length of characters and numbers.
        """
        return self._len(self.pwnb)

    @property
    def len_pun(self):
        """
        Length of all punctuations.
        """
        return self._len(self.ppun)

    @property
    def len_eng(self):
        """
        Length of English characters.
        """
        return self._len(self.peng)

    @property
    def len_num(self):
        """
        Length of all numbers.
        """
        return self._len(self.pnum)



import itertools
import re
from .ptxt import Regex
from .utils import pstr


psent = re.compile(r'''
    \n+
    |
    [。.！!?？…]+[”][。.!！?？…～~]?
    |
    (?<=[ \u3000a-zA-Z"”》）)〉〕】>」』\u4e00-\u9fa5])[。.!！?？…～~]+
    ''', re.UNICODE | re.VERBOSE)
psubsent = re.compile(r'''
    \n+
    |
    [。.！!?？…]+[”][。.!！?？…～~]?
    |
    (?<=[ \u3000a-zA-Z"”》）)〉〕】>」』\u4e00-\u9fa5])[,，、:：；;。.!！?？…～~]+
    ''', re.UNICODE | re.VERBOSE)
# referenced from jieba
punzh = pstr(Regex.pun_zh) - "-"  # for minus number eg -2
punen = pstr(Regex.pun_en) - "."  # for float number eg 1.3
pun = punzh + punen
pzh = re.compile(rf"([\u4E00-\u9FD5{pun}+#&])", re.UNICODE)
pen = re.compile(r"([a-zA-Z]+)", re.UNICODE)
pskip = re.compile(r"(\s)", re.UNICODE)
pspecial = re.compile(r"([-.])")  # split to single
pnum = re.compile(r"""
    ([-]?\d{1,}[.]?\d{0,}%)
    |
    ([-]?\d{1,}[./]?\d{0,})    
    """, re.UNICODE | re.VERBOSE)


def cut_zhchar(text: str, remove_blank: bool = False) -> list:
    lst = []
    blocks = pzh.split(text)
    for block in blocks:
        if not block:
            continue
        if pzh.match(block):
            for char in block:
                lst.append(char)
        else:
            skips = pskip.split(block)
            for skip in skips:
                if pen.search(skip):
                    for en_part in pen.split(skip):
                        if en_part:
                            spe = pspecial.search(en_part)
                            if not spe:
                                lst.append(en_part)
                            else:
                                for spe_part in pspecial.split(en_part):
                                    if spe_part:
                                        lst.append(spe_part)
                elif pnum.search(skip):
                    if skip[-1] != ".":
                        lst.append(skip)
                    else:
                        i = 0
                        while skip[-1] == ".":
                            i += 1
                            skip = skip[:-1]
                        lst.append(skip)
                        for _ in range(i):
                            lst.append(".")
                else:
                    if remove_blank:
                        skip = pskip.sub("", skip)
                    if skip:
                        lst.append(skip)
    return lst


def cut_part(
        text: str,
        split_pattern: re.Pattern,
        with_spliter: bool = True,
        with_offset: bool = False) -> list:
    """
    Cut text to parts by the given Regex Pattern.

    Parameters
    ----------
    text: raw text.
    split_pattern: how to split text.
    with_spliter: whether the parts contain spliters.
    with_offset: whether the parts contain offsets.

    Returns
    --------
    out: cutted parts.
    """
    spliters = split_pattern.findall(text)
    length = len(spliters)
    lst = []
    start = 0
    for i, part in enumerate(split_pattern.split(text)):
        if i < length:
            if with_spliter:
                part = part + spliters[i]
                len_spliter = 0
            else:
                len_spliter = len(spliters[i])
        else:
            len_spliter = 0
        end = start + len(part) + len_spliter
        if part:
            if with_offset:
                item = (part, start, end)
            else:
                item = part
            lst.append(item)
        start = end
    return lst


def cut_sentence(text: str) -> list:
    return cut_part(text, psent, True, False)


def combine_bucket(
        parts: list,
        threshold: int,
        truncate: bool = False,
        keep_remain: bool = False) -> list:
    """
    Convert parts to buckets with given length(threshold).

    Parameters
    ----------
    parts: the given parts.
    threshold: bucket length.
    truncate: whether to truncate those whose length is bigger than threshold.
    keep_remain: when truncate=True, whether to keep the remain parts.

    Returns
    out: list of bucket.
    -------
    """
    def deal_long_part(part: str) -> list:
        result = []
        if truncate:
            if keep_remain:
                len_subparts = len(part) // threshold + 1
                for i in range(len_subparts):
                    sub_part = part[i*threshold: (i+1)*threshold]
                    if sub_part:
                        result.append(sub_part)
            else:
                result.append(part[:threshold])
        else:
            result.append(part)
        return result

    buckets = []
    while parts:
        part = parts.pop(0)
        # directly add to buckets when a part is longer than threshold
        if len(part) > threshold:
            sub_parts = deal_long_part(part)
            buckets.append(sub_parts)
        else:
            while parts and len(part) < threshold:
                another = parts[0]
                if len(part + another) > threshold:
                    break
                else:
                    part += parts.pop(0)
            buckets.append([part])
    result = list(itertools.chain(*buckets))
    return result






import os
from pnlp.piop import read_lines

root = os.path.dirname(os.path.dirname(os.path.abspath(__file__)))


chinese_stopwords_file = os.path.join(root, "stopwords/chinese_stopwords.txt")
english_stopwords_file = os.path.join(root, "stopwords/english_stopwords.txt")


chinese_stopwords = set(read_lines(chinese_stopwords_file))
english_stopwords = set(read_lines(english_stopwords_file))


class StopWords:

    def __init__(self, path: str = ""):
        self._chinese_stopwords = chinese_stopwords
        self._english_stopwords = english_stopwords
        if path:
            self._stopwords = set(read_lines(path))
        else:
            self._stopwords = set()

    @property
    def zh(self):
        return self._chinese_stopwords

    @property
    def zh_len(self):
        return len(self._chinese_stopwords)

    @property
    def en(self):
        return self._english_stopwords

    @property
    def en_len(self):
        return len(self._english_stopwords)

    @property
    def stopwords(self):
        return self._stopwords



import copy
from dataclasses import dataclass, field
from typing import List, Dict, Callable, Tuple, Optional
from itertools import chain

import numpy as np

from .pcut import cut_zhchar, cut_part, psent, psubsent
from .ptxt import Regex
from .stopwords import chinese_stopwords, english_stopwords

reg = Regex()
STOPWORDS = list(english_stopwords | chinese_stopwords)
# 主要对功能词（与实词对应）采样处理
SAMPLE_WORDS = [w for w in STOPWORDS if reg.pwnb.search(w)]
# 副词、介词、连词、助词、其他虚词
SAMPLE_POS = ["d", "p", "c", "u", "xc"]


def swap(lst: list, index: int, start: int, end: int) -> list:
    """Randomly swap two continuous parts."""
    assert start <= index <= end <= len(lst) - 1
    if (index == start or np.random.rand() < 0.5) and index != end:
        lst[index], lst[index+1] = lst[index+1], lst[index]
    else:
        lst[index], lst[index-1] = lst[index-1], lst[index]
    return lst


@dataclass
class Sampler:

    def check_types(self):
        default_types = set(("delete", "swap", "insert"))
        for typ in self.types:
            if typ not in default_types:
                raise ValueError("Type {} is not a valid type.".format(typ))


@dataclass
class TokenLevelSampler(Sampler):

    """
    Random choose an index.
    - Insert a copy token. Usually a function word.
    - Delete a token.
    - Swap with the prev or the next token.

    Parameters
    -------------
    rate: The sampling rate (each type respectively).
    types: Sampling methods. You should take care of the order.
    sample_words: Words will be used in sampling. Usually use stopwords.
    sample_pos: Part-of-speech will be used in sampling. Usually use function pos.

    Note
    ------
    1. We mainly use `stopwords (usually function words)` as sample words to do the sampling.
    If you want to sample other kinds of words, you could appoint `sample_pos` to what you need,
    then your input should include POS flags.
    2. The order of the `types` will influence the output of the dependent_sample.
    """
    rate: float = 0.05
    types: List[str] = field(
        default_factory=lambda: ["delete", "swap", "insert"])
    sample_words: List[str] = field(
        default_factory=lambda: SAMPLE_WORDS)
    sample_pos: List[str] = field(
        default_factory=lambda: SAMPLE_POS)

    def __post_init__(self):
        self.len_types = len(self.types)
        self.check_types()
        assert self.rate <= 0.1

    def filter_sample_idx(
        self,
        token_list: List[str or Tuple[str, str]]
    ) -> List[int]:
        if not token_list:
            return []
        if type(token_list[0]) == str:
            can_deal_idx = [i for (i, w) in enumerate(
                token_list) if w in self.sample_words]
        else:
            can_deal_idx = [i for (i, (w, f)) in enumerate(
                token_list) if f in self.sample_pos]
        return can_deal_idx

    def choose_sample_idx(
        self,
        len_parts: int,
        sample_count: int
    ) -> List[int]:
        size = min(len_parts, sample_count)
        sample_part_idx = np.random.choice(
            len_parts, size, replace=False).tolist()
        return sample_part_idx

    def delete_sampling(
        self,
        token_list: List[str or Tuple[str, str]],
        sample_idx: List[int]
    ) -> List[str or Tuple[str, str]]:
        """Simple delete sampling. Delete the tokens in the given indexes.
        """
        result = []
        for i, token in enumerate(token_list):
            if i in sample_idx:
                continue
            result.append(token)
        return result

    def insert_sampling(
        self,
        token_list: List[str or Tuple[str, str]],
        sample_idx: List[int]
    ) -> List[str or Tuple[str, str]]:
        """Simple insert sampling. Insert the tokens in the given indexes.
        """
        result = []
        for i, token in enumerate(token_list):
            if i in sample_idx:
                result.append(token)
            result.append(token)
        return result

    def swap_sampling(
        self,
        token_list: List[str or Tuple[str, str]],
        sample_idx: List[int]
    ) -> List[str or Tuple[str, str]]:
        """Simple swap sampling. Swap the tokens in the given indexes. 
        DONOT swap start and end.
        """
        result = copy.deepcopy(token_list)
        end = len(token_list) - 1
        for idx in sample_idx:
            swap(result, idx, 0, end)
        return result

    def _sampling(
        self,
        type: str,
        parts:  List[List[str] or List[Tuple[str, str]]],
        sample_idx: List[int]
    ) -> List[List[str] or List[Tuple[str, str]]]:
        """
        Sampling by part, each time deal with a part instead of a token.
        """
        cp_parts = copy.deepcopy(parts)
        for j, part in enumerate(cp_parts):
            if j not in sample_idx:
                continue
            can_deal_idx = self.filter_sample_idx(part)
            if not can_deal_idx:
                continue
            deal_idx = np.random.choice(can_deal_idx, 1).tolist()[0]
            if type == "delete":
                part.remove(part[deal_idx])
            elif type == "insert":
                part.insert(deal_idx, part[deal_idx])
            else:
                swap(part, deal_idx, can_deal_idx[0], can_deal_idx[-1])
        return cp_parts

    def independent_sampling(
        self,
        token_list: List[str or Tuple[str, str]]
    ) -> List[List[str] or List[Tuple[str, str]]]:
        result = []
        parts = self.convert_tokens_to_parts_by_nonword(token_list)
        len_parts = len(parts)
        len_tokens = len(token_list)
        sample_count = round(len_tokens * self.rate) * self.len_types
        sample_count = max(sample_count, 1)
        # 一次采样到位，之后只是分别操作，操作原则上互相不依赖
        sample_part_idx = self.choose_sample_idx(len_parts, sample_count)
        each_count = len(sample_part_idx) // self.len_types

        for i, typ in enumerate(self.types):
            sample_idx = sample_part_idx[i*each_count: (i+1)*each_count]
            new_parts = self._sampling(typ, parts, sample_idx)
            sample = list(chain(*new_parts))
            result.append(sample)
        return result

    def dependent_sampling(
        self,
        token_list: List[str or Tuple[str, str]]
    ) -> List[str or Tuple[str, str]]:
        parts = self.convert_tokens_to_parts_by_nonword(token_list)
        len_parts = len(parts)
        len_tokens = len(token_list)
        sample_count = round(len_tokens * self.rate)
        sample_count = max(sample_count, 1)
        for i, typ in enumerate(self.types):
            # 每次重新采样，后面的操作可能会与前面的操作重叠
            sample_idx = self.choose_sample_idx(len_parts, sample_count)
            parts = self._sampling(typ, parts, sample_idx)
        return list(chain(*parts))

    def convert_tokens_to_parts_by_nonword(
        self,
        token_list: List[str or Tuple[str, str]]
    ) -> List[List[str] or List[Tuple[str, str]]]:
        parts = []
        tmp = []
        for token in token_list:
            tmp.append(token)
            word = self.__get_word_from_token(token)
            if reg.pnwn.search(word):
                parts.append(tmp)
                tmp = []
        return parts

    def __get_word_from_token(self, token: str or Tuple[str, str]) -> str:
        if type(token) == str:
            return token
        else:
            tup = tuple(token)
            return tup[0]

    def __join_tokens(self, token_list: List[str or Tuple[str, str]]) -> str:
        result = []
        for token in token_list:
            word = self.__get_word_from_token(token)
            result.append(word)
        return "".join(result)

    def make_samples(
        self,
        text: str,
        tokenizer: Optional[
            Callable[[str], List[str or Tuple[str, str]]]] = None
    ) -> Dict[str, str]:
        """
        Make negative samples.

        Parameters
        -----------
        text: The given text. Usually a sentence.
        tokenizer: Input a text, output a List of tokens. A token is a word or a (word, flag) tuple.

        Returns
        --------
        output: A dict of different kinds of negative samples.
        """
        if self.len_types == 0:
            return {}
        if not tokenizer:
            if reg.pchi.search(text):
                tokenizer = cut_zhchar
            else:
                def tokenizer(x): return x.split()
        tokens = tokenizer(text)
        if len(tokens) == 0:
            return {}
        result = {}
        indep_samples = self.independent_sampling(tokens)
        dep_sample = self.dependent_sampling(tokens)
        if indep_samples:
            for i, typ in enumerate(self.types):
                new_tokens = indep_samples[i]
                result[typ] = self.__join_tokens(new_tokens)
            result["together"] = self.__join_tokens(dep_sample)
        return result


@dataclass
class SentenceLevelSampler(Sampler):

    """
    Random choose an index. 
    - Insert a copy.
    - Delete.
    - Swap with the prev or the next one.

    We only deal with ONE sentence once.
    So you'd better use a paragraph as input.

    Parameters
    -----------
    types: Sampling methods. You should take care of the order.
    """

    types: List[str] = field(
        default_factory=lambda: ["delete", "swap", "insert"])

    def __post_init__(self):
        self.check_types()

    def independent_sampling(self, text_list: List[str]) -> List[List[str]]:
        result = []
        text_list = copy.deepcopy(text_list)
        length = len(text_list)
        for i, typ in enumerate(self.types):
            idx = np.random.choice(length, 1).tolist()[0]
            if typ == "insert":
                new = text_list[:idx] + [text_list[idx]] + text_list[idx:]
            elif typ == "delete":
                new = [s for (i, s) in enumerate(text_list) if i != idx]
            else:
                new = swap(text_list, idx, 0, length-1)
            result.append(new)
        return result

    def dependent_sampling(self, text_list: List[str]) -> List[str]:
        text_list = copy.deepcopy(text_list)
        for i, typ in enumerate(self.types):
            # 每次重新更新长度
            length = len(text_list)
            if length == 0:
                continue
            idx = np.random.choice(length, 1).tolist()[0]
            if typ == "insert":
                text_list = text_list[:idx] + \
                    [text_list[idx]] + text_list[idx:]
            elif typ == "delete":
                text_list = [s for (i, s) in enumerate(text_list) if i != idx]
            else:
                text_list = swap(text_list, idx, 0, length-1)
        return text_list

    def make_samples(self, text: str, level: str = "sent") -> List[str]:
        """
        Parameters
        -------------
        text: The given text. Always a paragraph.
        level: The sampling level. Could be one of {"sent", "subsent"}.

        Returns
        --------
        output: A dict of different kinds of negative samples.
        """
        if level == "sent":
            text_list = cut_part(text, psent)
        else:
            text_list = cut_part(text, psubsent)
        if len(text_list) == 0:
            return {}
        result = {}
        indep_samples = self.independent_sampling(text_list)
        dep_sample = self.dependent_sampling(text_list)
        if indep_samples:
            for i, typ in enumerate(self.types):
                new = indep_samples[i]
                result[typ] = "".join(new)
            result["together"] = "".join(dep_sample)
        return result





class pstr(str):
    def __sub__(self, other):
        result = []
        for c in self:
            if c in other:
                continue
            result.append(c)
        return "".join(result)




import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="pnlp",
    version="0.3.5",
    author="Yam",
    author_email="haoshaochun@gmail.com",
    description="A pre-processing tool for NLP.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/hscspring/pnlp",
    include_package_data=True,
    # default is `setup.py` path, so do not need a `package_dir` attr
    # if another dir, should be declared by `package_dir`
    packages=setuptools.find_packages(exclude=["*.tests", "*.tests.*", "tests.*", "tests"]),
    install_requires=[
        'addict',
        'pyyaml',
        'numpy'
    ],
    package_data={
        'pnlp': ["stopwords/*"],
    },
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: Apache Software License",
        "Operating System :: OS Independent",
    ],
)